/*
 * Copyright (c) 2005 Ichiro Satoh
 * All Rights Reserved.
 * written by Ichiro Satoh
 *
 * This program was developed by Ichiro Satoh at National Institute 
 * of Informatics as a part of a mobile object system, titled 
 * "AgentSpace".  All the programs of the system is developed by 
 * Ichiro Satoh and all rights are reserved by Ichiro Satoh. 
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the
 * copyright notice and this permission notice appear in all copies of
 * the software, derivative works or modified versions, and any
 * portions thereof, and that both notices appear in supporting 
 * documentation.
 */

/**
 * @author Ichiro Satoh
 * Created on 2005/05/5
 */

package agentspace.runtime;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.Vector;

import agentspace.Duplicatable;
import agentspace.IllegalAgentStateException;
import agentspace.InvalidAgentException;
import agentspace.Message;
import agentspace.Mobile;
import agentspace.Preservable;
import agentspace.Agent;
import agentspace.event.AgentEvent;
import agentspace.runtime.event.AgentLifecycleEvent;
import agentspace.runtime.event.AgentStatusEvent;
import agentspace.runtime.event.AgentStatusListener;
import agentspace.runtime.loader.MethodUtil;

public class AgentLifecycleManager implements Runnable {
	protected ThreadGroup theThreadGroup = null;
	protected AgentQueue queue = null;
	protected AgentContainer container = null;
	protected boolean active = false;

	public AgentLifecycleManager(AgentContainer container, ThreadGroup atg) {
		if (atg == null) {
			atg = new ThreadGroup("-AGENT");
		}
		theThreadGroup = atg;
		if (container == null) {
			container = new AgentContainer();
		}
		this.container = container;
		queue = new AgentQueue();
	}


	protected synchronized void setActive(boolean active) {
		this.active = active;
		notify();
	}

	protected synchronized boolean isActive() {
		return active;
	}
	
	public synchronized void processEvent(AgentLifecycleEvent evt)
		throws IllegalAgentStateException {
		if (evt.getAction() == AgentLifecycleEvent.HALT) {
			setActive(false);					
			setStatus(AgentStatus.LIFE_MASK, AgentStatus.HALT);
			halt();
			return;
		}
		while(!isActive()) {
			try {
				wait();
			} 
			catch (InterruptedException e) {
				e.printStackTrace();
				break;
			}
		}
//		System.err.println("AgentLifecycleManager: processEvent: "+ evt +" in "+AgentStatus.getLifeName(getStatus()));
		int status = getStatus() & AgentStatus.LIFE_MASK;
		switch (status) {
		case AgentStatus.UNINITIALIZED:
		case AgentStatus.SERIALIZED:
		case AgentStatus.CREATING:
		case AgentStatus.ARRIVING:
		case AgentStatus.NORMAL:
		case AgentStatus.UNSERIALIZING:
		case AgentStatus.DUPLICATING:
		case AgentStatus.DUPLICATED:
			if (isActive() && queue != null) {
				queue.enqueue(evt);
			} 
//		    System.err.println("AgentLifecycleManager: enqueued: "+ evt +" in "+AgentStatus.getLifeName(getStatus()));
		break;
		case AgentStatus.SERIALIZING:
		case AgentStatus.TERMINATING:
		case AgentStatus.TERMINATED:
		case AgentStatus.LEAVING:
		case AgentStatus.MOVING:
			throw new IllegalAgentStateException();
		case AgentStatus.HALT:
			halt();
		break;
		default:
			System.err.println("AgentLifecycleManager: " + evt.getActionName()+" in " + AgentStatus.getLifeName(getStatus()));
		}
	}

	protected void halt() {
		if (!Thread.currentThread().getThreadGroup().parentOf(theThreadGroup)) {
			theThreadGroup.interrupt();
			System.out.println("Halt: ac = "+container);
		}
	}

	public void run() {
//		System.out.println("AgentLifecycleManager: status = "+AgentStatus.getLifeName(getStatus()));
		setActive(true);
		setStatus(AgentStatus.PROCESS_MASK, AgentStatus.EXCLUSIVE);
		if ((getStatus() & AgentStatus.LIFE_MASK) == AgentStatus.UNINITIALIZED) {
			if (container.isAgent()) {
				setStatus(AgentStatus.LIFE_MASK, AgentStatus.CREATING);
				AgentEvent event = new AgentEvent(container.getObject());
				if (container.getObject() instanceof Agent) { 
					Agent agent = (Agent)container.getObject();
					agent.create(event, container.getContext());
				}
				else {
					throw new InvalidAgentException("Illeagal state : "+getStatus());
				}
			}
			else {
				throw new InvalidAgentException("Illeagal state : "+getStatus());
			}
		}
		else if ((getStatus() & AgentStatus.LIFE_MASK) == AgentStatus.MOVING) {				
			if (container.isMobileAgent()) { 
				setStatus(AgentStatus.LIFE_MASK, AgentStatus.ARRIVING);
				AgentEvent event = new AgentEvent(container.getObject());
				if (container.getObject() instanceof Mobile) { 
					Mobile agent = (Mobile)container.getObject();
					agent.arrive(event, container.getContext());
				}
			}
		}
		else if ((getStatus() & AgentStatus.LIFE_MASK) == AgentStatus.SERIALIZED) {				
			if (container.isPreservableAgent()) { 
				setStatus(AgentStatus.LIFE_MASK, AgentStatus.UNSERIALIZING);
				AgentEvent event = new AgentEvent(container.getObject());
				if (container.getObject() instanceof Preservable) { 
					Preservable agent = (Preservable)container.getObject();
					agent.resume(event, container.getContext());
				}
			}
		}
		else if ((getStatus() & AgentStatus.LIFE_MASK) == AgentStatus.DUPLICATED) {				
			if (container.isDuplicatableAgent()) { 
				AgentEvent event = new AgentEvent(container.getObject());
				if (container.getObject() instanceof Duplicatable) { 
					Duplicatable agent = (Duplicatable)container.getObject();
					agent.child(event, container.getContext());
				}
			}
		}
		else if ((getStatus() & AgentStatus.LIFE_MASK) == AgentStatus.CHILD) {				
			if (container.isDuplicatableAgent()) { 
				AgentEvent event = new AgentEvent(container.getObject());
				if (container.getObject() instanceof Duplicatable) { 
					Duplicatable agent = (Duplicatable)container.getObject();
					agent.child(event, container.getContext());
				}
			}
		}
		else {
			throw new InvalidAgentException("Illeagal state : "+getStatus());
		}
		setStatus(AgentStatus.LIFE_MASK, AgentStatus.NORMAL);
		setStatus(AgentStatus.PROCESS_MASK, AgentStatus.READY);
		AgentLifecycleEvent evt = null;
		while (true) {
			try {
				evt = (AgentLifecycleEvent) queue.dequeue();
				if (evt.getAction() == AgentLifecycleEvent.SUSPEND) {
					setActive(false);					
					setStatus(AgentStatus.LIFE_MASK, AgentStatus.SERIALIZING);
					break;
				}
				else if (evt.getAction() == AgentLifecycleEvent.LEAVE) {
					setActive(false);					
					setStatus(AgentStatus.LIFE_MASK, AgentStatus.LEAVING);
					break;
				}
				else if (evt.getAction() == AgentLifecycleEvent.DESTROY) {
					setActive(false);					
					setStatus(AgentStatus.LIFE_MASK, AgentStatus.TERMINATING);
					break;
				}
				else if (evt.getAction() == AgentLifecycleEvent.DUPLICATE) {
					setActive(false);					
					setStatus(AgentStatus.LIFE_MASK, AgentStatus.DUPLICATING);
					break;
				}
				else if (evt.getAction() == AgentLifecycleEvent.MESSAGE) {
					setStatus(AgentStatus.PROCESS_MASK, AgentStatus.BUSY);
					Message msg = evt.getMessage();
//					System.out.println("AgentLifecycleManager: msg = "+msg);
					if (!isValidMessage(msg)) {
						evt.setException(new NoSuchMethodException());
					}
					else {
						try {
							AgentMessageHandler handler = callback(msg, false);
							evt.setFuture(msg.getFuture());
						}
						catch (Exception ex) {
							evt.setException(ex);
							ex.printStackTrace();
						}
					}
					evt.setFinished(true);
					setStatus(AgentStatus.PROCESS_MASK, AgentStatus.READY);
				}
				else if (evt.getAction() == AgentLifecycleEvent.HALT) {
					setActive(false);					
					setStatus(AgentStatus.LIFE_MASK, AgentStatus.HALT);
					break;
				}
				else {
					continue;
				}
			}
			catch (InterruptedException ex) {
				ex.printStackTrace();
				break;
			} 
		}
		setStatus(AgentStatus.PROCESS_MASK, AgentStatus.EXCLUSIVE);
		if ((getStatus() & AgentStatus.LIFE_MASK) == AgentStatus.LEAVING) {
			if (container.isMobileAgent()) { 
				setStatus(AgentStatus.LIFE_MASK, AgentStatus.LEAVING);
				AgentEvent event = new AgentEvent(container.getObject());
				if (container.getObject() instanceof Mobile) { 
					Mobile agent = (Mobile)container.getObject();
					agent.leave(event, container.getContext());
				}
			}
			setStatus(AgentStatus.LIFE_MASK, AgentStatus.MOVING);
		}
		else if ((getStatus() & AgentStatus.LIFE_MASK) == AgentStatus.SERIALIZING) {
			if (container.isPreservableAgent()) { 
				setStatus(AgentStatus.LIFE_MASK, AgentStatus.SERIALIZING);
				AgentEvent event = new AgentEvent(container.getObject());
				if (container.getObject() instanceof Preservable) { 
					Preservable agent = (Preservable)container.getObject();
					agent.suspend(event, container.getContext());
				}
			}
			setStatus(AgentStatus.LIFE_MASK, AgentStatus.SERIALIZED);
		}
		else if ((getStatus() & AgentStatus.LIFE_MASK) == AgentStatus.TERMINATING) {
			if (container.isAgent()) {
				setStatus(AgentStatus.LIFE_MASK, AgentStatus.TERMINATING);
				AgentEvent event = new AgentEvent(container.getObject());
				if (container.getObject() instanceof Agent) { 
					Agent agent = (Agent)container.getObject();
					agent.destroy(event, container.getContext());
				}
			}
			setStatus(AgentStatus.LIFE_MASK, AgentStatus.TERMINATED);
		}
		else if ((getStatus() & AgentStatus.LIFE_MASK) == AgentStatus.DUPLICATING) {
			if (container.isDuplicatableAgent()) { 
				AgentEvent event = new AgentEvent(container.getObject());
				if (container.getObject() instanceof Duplicatable) { 
					Duplicatable agent = (Duplicatable)container.getObject();
					agent.duplicate(event, container.getContext());
				}
			}
			setStatus(AgentStatus.LIFE_MASK, AgentStatus.DUPLICATED);
		}
		else {
			setStatus(AgentStatus.LIFE_MASK, AgentStatus.EXCEPTION);
		}
		evt.setFinished(true);
	}

	protected boolean isValidMessage(Message msg) {
		try {
			Method method = MethodUtil.getMethod(container.getObject().getClass(), 
					msg.getName(), msg.getTypes(), false);
			if (method != null) {
				return true;
			}
		} 
		catch (NoSuchMethodException e) {}
		return false;
	}
	
	protected AgentMessageHandler callback(Message msg, boolean sync) {
		AgentMessageHandler handler = new AgentMessageHandler(container.getObject(), msg);
		if (sync) {
			handler.run();
		}
		else {
			Thread t = new Thread(handler);
			t.start();
		}
		return handler;
	}

	
	protected int getStatus() {
		return container.getStatus();
	}
	
	protected synchronized void setStatus(int mask, int status) {
		container.setStatus((container.getStatus() & ~mask) | status);
		//System.err.println("AgentLifecycleManager: " + AgentStatus.getLifeName(getStatus())+", "+AgentStatus.getProcessName(getStatus()));
		if (statusListeners != null) {
			AgentStatusEvent statusEvent = new AgentStatusEvent(container, status, mask);
			Enumeration enum = (Enumeration)statusListeners.elements();
			while(enum.hasMoreElements()) {
				AgentStatusListener listener = (AgentStatusListener)enum.nextElement();
		    	listener.agentChanged(statusEvent);
			}
		}
	}
		
	protected Vector statusListeners = null; 
	
	public synchronized void addAgentStatusListener(AgentStatusListener listener) {
		if (statusListeners == null) {
			statusListeners = new Vector();
		}
		statusListeners.addElement(listener);	
	}
	
	public synchronized void removeAgentStatusListener(AgentStatusListener listener) {
		if (statusListeners != null) {
			statusListeners.removeElement(listener);
		}
		if (statusListeners.size() == 0) {
			statusListeners.clear();
			statusListeners = null;
		}
	} 
}
